// Code to load symbols from a legacy filesystem
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <cstdio>
#include <vector>
#include <memory>
#include <cstring>

#include <errno.h>
#include <dirent.h>

#include <idlc.h>
#include <parser.h>

// To deal with namespaces on a case-insensitive filesystem, names
// are annotated with a hexadecimal prefix indicating the case of each
// letter.  Each bit in the number represents a letter, with 0 being
// lowercase (or a non-letter) and 1 being uppercase.  The
// least-significant bit is the first letter.  This is done on both
// case sensitive and case insensitive filesystems so that the output
// can be freely copied between them (it also simplifies the code).
// As such, hex digits should be in upper case so that case-sensitive
// lookups can be used.  This prefix is separated from the actual name
// by an underscore.
//
// When the time comes along that UTF-8 is supported, this applies to
// UTF-8 characters, not ASCII characters.
//
// This mangling can also be used (with a leading underscore or
// whatever) in case-insensitive languages, though ideally there
// should be aliases with the unannotated name if there is no
// conflict.
//
// I don't want to go the CORBA route of disallowing case conflicts,
// as I find using capitals for types and lowercase for instances to
// be a useful convention (and less ugly than appending _t, except
// when case sensitivity is missing), and would rather not be limited
// at such a fundamental level by broken languages.

static char tohex(int digit)
{
	if (digit < 10)
		return digit + '0';
	
	return digit + 'A' - 10;
}

const String *NameSpace::mangle(const String *name)
{
	String *mangled_string = new String;
	
	// The conversion is done manually so that I don't have to deal
	// with bignum arithmetic.
	
	std::vector<bool> hex;
	std::vector<bool>::size_type name_len = name->length();
	
	// Pad out the first digit with leading zeroes, so that
	// we don't have to deal with that later.
	
	if (name_len & 3)
		for (int i = name_len & 3; i < 4; ++i)
			hex.push_back(0);

	for (string::const_iterator i = name->begin(); i != name->end(); ++i)
		hex.push_back(isupper(*i));

	assert((hex.size() & 3) == 0);

	for (std::vector<bool>::size_type i = 0; i < hex.size(); i += 4) {
		int digit = hex[i]     * 8 +
		            hex[i + 1] * 4 +
		            hex[i + 2] * 2 +
		            hex[i + 3];

		*mangled_string += tohex(digit);
	}
	
	*mangled_string += '_';
	*mangled_string += *name;
	return mangled_string;
}

struct IOError
{
	const char *msg;

	IOError(FILE *f)
	{
		if (ferror(f))
			msg = strerror(ferror(f));
		else
			msg = "Short File";
	}
};

struct ImportContext {
	const String *name;
	const char *filename;
	File f;
	NameSpace *parent;
	bool swap;
	bool is_dir;
	bool is_typedef;
	CompiledDefHeader hdr;
};

static void import_and_validate_array(ImportContext &ctx,
                                      CompiledArray &out,
                                      CompiledArray &in)
{
	out.bounds[0] = swap64(in.bounds[0], ctx.swap);
	out.bounds[1] = swap64(in.bounds[1], ctx.swap);

	if (out.bounds[0] < 0) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(bad array lower bound %" PRId64 ").\n",
		        ctx.filename, out.bounds[0]);
		throw UserError();
	}
	
	if (out.bounds[1] < -1) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(bad array upper bound %" PRId64 ").\n",
		        ctx.filename, out.bounds[0]);
		throw UserError();
	}
	
	if (out.bounds[1] > 0 &&
	    out.bounds[1] < out.bounds[0]) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(array upper bound %" PRId64 " less than lower bound %" PRId64 ").\n",
		        ctx.filename, out.bounds[1], out.bounds[0]);
		throw UserError();
	}
}

static void import_and_validate_basictype(ImportContext &ctx,
                                          CompiledBasicType &out, 
                                          CompiledBasicType &in)
{
	out.bits = swap32(in.bits, ctx.swap);
	out.flags.raw = swap32(in.flags.raw, ctx.swap);

	int num = 0;
	if (out.flags.field.Unsigned)
		num++;
	if (out.flags.field.Float)
		num++;
	if (out.flags.field.Bool)
		num++;
	if (num > 1) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(bad flags 0x%08x).\n",
		        ctx.filename, out.flags.raw);
		throw UserError();
	}

	if (out.flags.field.Bool) {
		if (out.bits != 0) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(bad bits %d for bool type).\n",
			        ctx.filename, out.bits);
			throw UserError();
		}
	} else if (out.flags.field.Float) {
		if (out.bits != 32 && out.bits != 64) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(bad bits %d for floating point type).\n",
			        ctx.filename, out.bits);
			throw UserError();
		}
	} else if (!dynamic_cast<BitField *>(ctx.parent) &&
	           !dynamic_cast<Enum *>(ctx.parent)) {
		if (out.bits != 8 && out.bits != 16 &&
		    out.bits != 32 && out.bits != 64) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(bad bits %d for integer type).\n",
			        ctx.filename, out.bits);
			throw UserError();
		}
		
		if (out.bits == 8 && !out.flags.field.Unsigned) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(octet type must be unsigned).\n",
			        ctx.filename);
			throw UserError();
		}
	}
	
	// Bitfield entries can take anywhere from 1 to 64 bits, but
	// will be verified by the parent after all entries have loaded,
	// so it can verify that the sum does not exceed the size of
	// the bitfield.

	import_and_validate_array(ctx, out.array, in.array);
}

BasicType *BasicType::import(ImportContext &ctx)
{
	if (ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(a BasicType must be a file).\n",
		        ctx.filename);
		throw UserError();
	}

	CompiledBasicType def;
	if (fread(&def, sizeof(def), 1, ctx.f) != 1)
		throw IOError(ctx.f);

	BasicType *bt = new BasicType(ctx.name);
	ctx.parent->add_import(bt, ctx.filename);
	
	import_and_validate_basictype(ctx, bt->def, def);

	bt->complete = true;
	return bt;
}

static String *read_string_raw(ImportContext &ctx, int32_t length)
{
	if (length < 0 || length > 4096) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(unreasonable string length %d).\n",
		        ctx.filename, length);
		throw UserError();
	}

	// Is there a way to reserve space in a C++ string, and pass
	// the buffer into fread?
	
	char *buf = new char[length + 1];
	
	if (fread(buf, length + 1, 1, ctx.f) != 1)
		throw IOError(ctx.f);
	
	String *s = new String(buf);
	delete[] buf;
	return s;
}

static const String *read_string(ImportContext &ctx)
{
	off_t pos = ftello(ctx.f);
	int pad = ((pos + 3) & ~3) - pos;

	if (pad != 0 && fseeko(ctx.f, pad, SEEK_CUR) != 0)
		throw IOError(ctx.f);

	int32_t length;
	if (fread(&length, sizeof(length), 1, ctx.f) != 1)
		throw IOError(ctx.f);
	
	return read_string_raw(ctx, swap32(length, ctx.swap));
}

StrList::StrList(const String *input, char delimeter)
{
	const char *cur = input->c_str();
	
	do {
		char *next = strchr(cur, delimeter);
		String *s = new String();
		s->token = TOK_IDENT;
		
		if (next == cur || *cur == 0)
			throw InvalidArgument();
		
		if (next) {
			s->append(cur, next - cur);
			cur = next + 1;
		} else {
			s->append(cur);
			cur = NULL;
		}
		
		push_back(s);
	} while (cur);
}

// FIXME: Handle illegal recursion.
static Symbol *lookup_sym_noyyerror(ImportContext &ctx, const String *name)
{
	StrList *sl;

	try {
		sl = new StrList(name);
	}

	catch (InvalidArgument) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(\"%s\" is not a valid identifier).\n",
		        ctx.filename, name->c_str());
		throw UserError();
	}

	try {
		return lookup_sym(toplevel, sl, toplevel);
	}
	
	catch (UserError) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(\"%s\" could not be loaded).\n",
		        ctx.filename, name->c_str());

		throw UserError();
	}
}

// FIXME: Handle illegal recursion.
static Symbol *lookup_sym_in_ns_noyyerror(ImportContext &ctx,
                                          NameSpace *ns,
                                          const String *name)
{
	Symbol *sym = ns->lookup_noex(name);
	if (!sym) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(\"%s\" is not found).\n",
		        ctx.filename, name->c_str());

		throw UserError();
	}
	
	return sym;
}

Datum *Datum::import(ImportContext &ctx)
{
	if (ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(a Datum must be a file).\n",
		        ctx.filename);
		throw UserError();
	}

	CompiledDatum def;
	if (fread(&def, sizeof(def), 1, ctx.f) != 1)
		throw IOError(ctx.f);

	Datum *d = new Datum(ctx.name);
	d->def.flags.raw = swap32(def.flags.raw, ctx.swap);

	if (d->def.flags.field.Invalid) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(bad flags 0x%08x).\n",
		        ctx.filename, d->def.flags.raw);
		throw UserError();
	}

	ctx.parent->add_import(d, ctx.filename);

	d->def.type.length = swap32(def.type.length, ctx.swap);

	if (d->def.type.length != 0) {
		d->type_fq_name = read_string_raw(ctx, d->def.type.length);
		Symbol *sym = lookup_sym_noyyerror(ctx, d->type_fq_name);
		d->type = dynamic_cast<Type *>(sym);
		
		if (!d->type) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(\"%s\" is not a type).\n",
			        ctx.filename, d->type_fq_name->c_str());
			throw UserError();
		}
		
		import_and_validate_array(ctx, d->def.basictype.array,
		                          def.basictype.array);
	
		BasicType *bt = dynamic_cast<BasicType *>(*d->type);
		if (bt)
			d->cbt = &bt->def;
		else
			d->cbt = NULL;
	} else {
		d->basic = true;
		import_and_validate_basictype(ctx, d->def.basictype,
		                              def.basictype);
		d->cbt = &d->def.basictype;
	}

	d->def.ucon = swap64(def.ucon, ctx.swap);

	if (d->def.flags.field.Const) {
		if (!d->cbt) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(constant, but non-basic type).\n",
			        ctx.filename);
			throw UserError();
		}
	
		if (d->cbt->flags.field.Float)
			d->con_type = TOK_FCON;
		else if (!d->cbt->flags.field.Unsigned)
			d->con_type = TOK_ICON;
		else
			d->con_type = TOK_UCON;
	
		if (!d->verify_const()) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(bad constant).\n",
			        ctx.filename);
			throw UserError();
		}
	} else {
		d->con_type = TOK_NONE;
	}

	return d;
}

// OPT: Importing methods and supers isn't necessary for idlc to
// work; it could be made optional to speed up large compilations.

Interface *Interface::import(ImportContext &ctx)
{
	if (!ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(an Interface must be a directory).\n",
		        ctx.filename);
		throw UserError();
	}

	CompiledInterface def;
	if (fread(&def, sizeof(def), 1, ctx.f) != 1)
		throw IOError(ctx.f);

	Interface *iface = new Interface(ctx.name);
	iface->path = new String(ctx.filename);
	ctx.parent->add_import(iface, ctx.filename);
	
	iface->def.guid[0] = def.guid[0];
	iface->def.guid[1] = def.guid[1];

	int32_t num_methods = swap32(def.num_methods, ctx.swap);
	int32_t num_supers = swap32(def.num_supers, ctx.swap);
	
	if (num_methods < 0 || num_methods > 4096) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(unreasonable num_methods %d).\n",
		        ctx.filename, num_methods);
		throw UserError();
	}

	if (num_supers < 0 || num_supers > 4096) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(unreasonable num_supers %d).\n",
		        ctx.filename, num_supers);
		throw UserError();
	}
	
	for (int32_t i = 0; i < num_methods; i++) {
		const String *str = read_string(ctx);
		Symbol *sym = lookup_sym_in_ns_noyyerror(ctx, iface, str);
		Method *m = dynamic_cast<Method *>(sym);
		
		if (!m) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(\"%s\" is not a method).\n",
			        ctx.filename, str->c_str());
			throw UserError();
		}
		
		iface->add_elem(m);
	}

	// FIXME: Check for bad recursion again

	for (int32_t i = 0; i < num_supers; i++) {
		const String *str = read_string(ctx);
		Symbol *sym = lookup_sym_noyyerror(ctx, str);
		Interface *super = dynamic_cast<Interface *>(sym);
	
		if (!super) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(\"%s\" is not an interface).\n",
			        ctx.filename, str->c_str());
			throw UserError();
		}
	
		iface->add_super(super);
	}
	
	assert(num_methods == iface->def.num_methods);
	assert(num_supers == iface->def.num_supers);

	iface->sort_chains();
	return iface;
}

Method *Method::import(ImportContext &ctx)
{
	if (!ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(a Method must be a directory).\n",
		        ctx.filename);
		throw UserError();
	}

	CompiledMethod def;
	if (fread(&def, sizeof(def), 1, ctx.f) != 1)
		throw IOError(ctx.f);

	Method *m = new Method(ctx.name);
	m->path = new String(ctx.filename);
	ctx.parent->add_import(m, ctx.filename);

	int32_t num_entries = swap32(def.num_entries, ctx.swap);

	m->def.flags.raw = swap32(def.flags.raw, ctx.swap);
	
	if (num_entries < 0 || num_entries > 4096) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(unreasonable num_entries %d).\n",
		        ctx.filename, num_entries);
		throw UserError();
	}

	for (int32_t i = 0; i < num_entries; i++) {
		const String *str = read_string(ctx);
		Symbol *sym = lookup_sym_in_ns_noyyerror(ctx, m, str);
		Param *p = dynamic_cast<Param *>(sym);
		
		if (!p) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(\"%s\" is not a parameter).\n",
			        ctx.filename, str->c_str());
			throw UserError();
		}
		
		m->add_elem(p);
	}

	assert(num_entries == m->def.num_entries);
	return m;
}

Param *Param::import(ImportContext &ctx)
{
	if (ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(a Parameter must be a file).\n",
		        ctx.filename);
		throw UserError();
	}

	CompiledParam def;
	if (fread(&def, sizeof(def), 1, ctx.f) != 1)
		throw IOError(ctx.f);

	Param *p = new Param(ctx.name);
	ctx.parent->add_import(p, ctx.filename);

	p->def.type.length = swap32(def.type.length, ctx.swap);

	if (p->def.type.length != 0) {
		p->type_fq_name = read_string_raw(ctx, p->def.type.length);
		Symbol *sym = lookup_sym_noyyerror(ctx, p->type_fq_name);
		p->type = dynamic_cast<Type *>(sym);
		
		if (!p->type) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(\"%s\" is not a type).\n",
			        ctx.filename, p->type_fq_name->c_str());
			throw UserError();
		}
		
		import_and_validate_array(ctx, p->def.basictype.array,
		                          def.basictype.array);
	} else {
		import_and_validate_basictype(ctx, p->def.basictype,
		                              def.basictype);
	}

	p->def.flags.raw = swap32(def.flags.raw, ctx.swap);
	return p;
}

// OPT: Importing data and super isn't necessary for idlc to
// work; it could be made optional to speed up large compilations.

Struct *Struct::import(ImportContext &ctx)
{
	if (!ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(a Struct must be a directory).\n",
		        ctx.filename);
		throw UserError();
	}

	CompiledStruct def;
	if (fread(&def, sizeof(def), 1, ctx.f) != 1)
		throw IOError(ctx.f);

	Struct *s = new Struct(ctx.name);
	s->path = new String(ctx.filename);
	ctx.parent->add_import(s, ctx.filename);
	
	s->def.guid[0] = def.guid[0];
	s->def.guid[1] = def.guid[1];
 	
	int32_t num_entries = swap32(def.num_entries, ctx.swap);

	s->def.flags.raw = swap32(def.flags.raw, ctx.swap);
	
	if (num_entries < 0 || num_entries > 4096) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(unreasonable num_entries %d).\n",
		        ctx.filename, num_entries);
		throw UserError();
	}

	for (int32_t i = 0; i < num_entries; i++) {
		const String *str = read_string(ctx);
		
		Symbol *sym = lookup_sym_in_ns_noyyerror(ctx, s, str);
		Datum *d = dynamic_cast<Datum *>(sym);
		
		if (!d) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(\"%s\" is not a datum).\n",
			        ctx.filename, str->c_str());
			throw UserError();
		}
		
		s->add_elem(d);
	}
	
	if (s->def.flags.field.Super) {
		const String *str = read_string(ctx);
		Symbol *sym = lookup_sym_noyyerror(ctx, str);
		Struct *super = dynamic_cast<Struct *>(sym);
		
		if (!super) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(\"%s\" is not a struct).\n",
			        ctx.filename, str->c_str());
			throw UserError();
		}
		
		s->super = super;
		
		if (super->is_virtual() && !s->is_virtual()) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(not virtual but parent is).\n",
			        ctx.filename);
			throw UserError();
		}
	}

	assert(num_entries == s->def.num_entries);
	return s;
}

// OPT: Importing elements isn't necessary for idlc to work (at
// least, not unless inheritance is implemented); it could be made
// optional to speed up large compilations.

BitField *BitField::import(ImportContext &ctx)
{
	if (!ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(a BitField must be a directory).\n",
		        ctx.filename);
		throw UserError();
	}

	CompiledBitField def;
	if (fread(&def, sizeof(def), 1, ctx.f) != 1)
		throw IOError(ctx.f);

	BitField *bf = new BitField(ctx.name);
	bf->path = new String(ctx.filename);
	ctx.parent->add_import(bf, ctx.filename);

	int32_t num_entries = swap32(def.num_entries, ctx.swap);
	bf->def.bits = swap32(def.bits, ctx.swap);

	if (num_entries < 0 || num_entries > 4096) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(unreasonable num_entries %d).\n",
		        ctx.filename, num_entries);
		throw UserError();
	}
	
	// FIXME: bits can only be 16, 32, or 64 when not in another bitfield.

	if (bf->def.bits < 1 || bf->def.bits > 64) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(unreasonable bits %d).\n",
		        ctx.filename, num_entries);
		throw UserError();
	}

	for (int32_t i = 0; i < num_entries; i++) {
		const String *str = read_string(ctx);
		Symbol *sym = lookup_sym_in_ns_noyyerror(ctx, bf, str);
		Datum *d = dynamic_cast<Datum *>(sym);
		
		if (!d) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(\"%s\" is not a datum).\n",
			        ctx.filename, str->c_str());
			throw UserError();
		}
		
		bf->add_elem(d);
	}
	
	assert(num_entries == bf->def.num_entries);
	return bf;
}

// OPT: Importing elements isn't necessary for idlc to work (at
// least, not unless inheritance is implemented); it could be made
// optional to speed up large compilations.

Enum *Enum::import(ImportContext &ctx)
{
	if (!ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(a Enum must be a directory).\n",
		        ctx.filename);
		throw UserError();
	}

	CompiledEnum def;
	if (fread(&def, sizeof(def), 1, ctx.f) != 1)
		throw IOError(ctx.f);

	Enum *e = new Enum(ctx.name);
	e->path = new String(ctx.filename);
	ctx.parent->add_import(e, ctx.filename);

	int32_t num_entries = swap32(def.num_entries, ctx.swap);
	e->def.bits = swap32(def.bits, ctx.swap);

	if (num_entries < 0 || num_entries > 4096) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(unreasonable num_entries %d).\n",
		        ctx.filename, num_entries);
		throw UserError();
	}
	
	// FIXME: bits can only be 16, 32, or 64 when not in another bitfield.

	if (e->def.bits < 1 || e->def.bits > 64) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(unreasonable bits %d).\n",
		        ctx.filename, num_entries);
		throw UserError();
	}

	for (int32_t i = 0; i < num_entries; i++) {
		const String *str = read_string(ctx);
		Symbol *sym = lookup_sym_in_ns_noyyerror(ctx, e, str);
		Datum *d = dynamic_cast<Datum *>(sym);
		
		if (!d) {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(\"%s\" is not a datum).\n",
			        ctx.filename, str->c_str());
			throw UserError();
		}
		
		e->add_elem(d);
	}
	
	assert(num_entries == e->def.num_entries);
	return e;
}

Alias *Alias::import(ImportContext &ctx)
{
	if (ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(an Alias must be a file).\n",
		        ctx.filename);
		throw UserError();
	}

	Alias *a = new Alias(ctx.name);
	ctx.parent->add_import(a, ctx.filename);

	const String *str = read_string(ctx);
	Symbol *sym = lookup_sym_noyyerror(ctx, str);
	
	Alias *aptr = dynamic_cast<Alias *>(sym);
	if (aptr) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(points to \"%s\", which is an alias).\n",
		        ctx.filename, str->c_str());
		throw UserError();
	}

	a->real_sym = sym;
	a->sym_fq_name = str;
	a->def.length = str->length();

	return a;
}

TypeDef *TypeDef::import(ImportContext &ctx)
{
	if (ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(a TypeDef must be a file).\n",
		        ctx.filename);
		throw UserError();
	}

	TypeDef *td = new TypeDef(ctx.name);
	ctx.parent->add_import(td, ctx.filename);

	const String *str = read_string(ctx);
	Symbol *sym = lookup_sym_noyyerror(ctx, str);
	
	Alias *aptr = dynamic_cast<Alias *>(sym);
	if (aptr) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(points to \"%s\", which is an alias).\n",
		        ctx.filename, str->c_str());
		throw UserError();
	}

	td->real_sym = sym;
	td->sym_fq_name = str;
	td->def.length = str->length();

	return td;
}

UserNameSpace *UserNameSpace::import(ImportContext &ctx)
{
	if (!ctx.is_dir) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(a NameSpace must be a directory).\n",
		        ctx.filename);
		throw UserError();
	}

	UserNameSpace *uns = new UserNameSpace();
	uns->path = new String(ctx.filename);
	
	// FIXME: sanity check the mount point

	uns->name = ctx.name;	
	ctx.parent->add_import(uns, ctx.filename);
	return uns;
}

Symbol *do_load(ImportContext &ctx)
{
	int type = swap32(ctx.hdr.type, ctx.swap);
	
	switch (type) {
		case CompiledDefHeader::NameSpace:
			return UserNameSpace::import(ctx);
			break;
		
		case CompiledDefHeader::BasicType:
			return BasicType::import(ctx);
			break;

		case CompiledDefHeader::Datum:
			return Datum::import(ctx);
			break;

		case CompiledDefHeader::Interface:
			return Interface::import(ctx);
			break;

		case CompiledDefHeader::Method:
			return Method::import(ctx);
			break;

		case CompiledDefHeader::Param:
			return Param::import(ctx);
			break;

		case CompiledDefHeader::Struct:
			return Struct::import(ctx);
			break;

		case CompiledDefHeader::Enum:
			return Enum::import(ctx);
			break;

		case CompiledDefHeader::BitField:
			return BitField::import(ctx);
			break;

		case CompiledDefHeader::Alias:
			ctx.is_typedef = false;
			return Alias::import(ctx);
			break;

		case CompiledDefHeader::TypeDef:
			ctx.is_typedef = true;
			return TypeDef::import(ctx);
			break;

		default:
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(bad type %08x).\n",
			        ctx.filename, type);
			
			throw UserError();
	}
}

Symbol *NameSpace::load(const String *symname)
{
	ImportContext ctx;
	const String *mangled = mangle(symname);

	ctx.name = new String(*symname);
	ctx.is_dir = true;
	ctx.parent = this;

	string filename_no_self(path);
	filename_no_self += '/';
	filename_no_self += *mangled;
	ctx.filename = filename_no_self.c_str();

	string filename(filename_no_self);
	filename.append("/.self");

	ctx.f = fopen(filename.c_str(), "rb");
	if (!ctx.f) {
		ctx.is_dir = false;
		ctx.f = fopen(ctx.filename, "rb");
		
		if (!ctx.f)
			return NULL;
		
		filename = filename_no_self;
	}
	
	Symbol *ret;
	try {
		if (fread(&ctx.hdr, sizeof(ctx.hdr), 1, ctx.f) != 1)
			throw IOError(ctx.f);

		if (ctx.hdr.magic == CompiledDefHeader::magic_normal)
			ctx.swap = false;
		else if (ctx.hdr.magic == CompiledDefHeader::magic_reversed)
			ctx.swap = true;
		else {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(bad magic 0x%08x).\n",
			        ctx.filename, ctx.hdr.magic);
	
			throw UserError();
		}
		
		ret = do_load(ctx);
	}
	
	catch (IOError &e) {
		fprintf(stderr, "Cannot read from file \"%s\": %s.\n",
		        ctx.filename, e.msg);
	
		throw UserError();
	}
	
	return ret;
}

NameSpace *check_for_imports(NameSpace *ns)
{
	while (ns) {
		if (ns->get_path())
			return ns;

		ns = ns->get_ns();
	}
	
	return NULL;
}

void UserNameSpace::declare_import(const char *path)
{
	ImportContext ctx;
	UserNameSpace *ns = new UserNameSpace();
	string filename(path);
	filename.append("/.self");

	ctx.parent = NULL;
	ctx.filename = path;
	
	ctx.f = fopen(filename.c_str(), "rb");
	if (!ctx.f) {
		fprintf(stderr, "Cannot import \"%s\": %s.\n",
		        path, strerror(errno));
	
		throw UserError();
	}
	
	const String *mount;
	try {
		if (fread(&ctx.hdr, sizeof(ctx.hdr), 1, ctx.f) != 1)
			throw IOError(ctx.f);
	
		if (ctx.hdr.magic == CompiledDefHeader::magic_normal)
			ctx.swap = false;
		else if (ctx.hdr.magic == CompiledDefHeader::magic_reversed)
			ctx.swap = true;
		else {
			fprintf(stderr, "\"%s\" is not a valid object "
			                "(bad magic 0x%08x).\n",
			        ctx.filename, ctx.hdr.magic);
	
			throw UserError();
		}
		
		mount = read_string(ctx);
	}
	
	catch (IOError &e) {
		fprintf(stderr, "Cannot read from file \"%s\": %s.\n",
		        ctx.filename, e.msg);
	
		throw UserError();
	}

	StrList *strl = new StrList(mount);
	if (!strl) {
		fprintf(stderr, "\"%s\" is not a valid object "
		                "(mount point \"%s\" is not valid).\n",
		        ctx.filename, mount->c_str());

		throw UserError();
	}

	ns->name = strl->back();
	strl->pop_back();
	
	if (strl->size() != 0) {
		ctx.parent = add_nspace(strl, false);

		if (!ctx.parent)
			BUG();
		
		NameSpace *conflict = check_for_imports(ctx.parent);
		if (conflict) {		
			fprintf(stderr, "Import \"%s\" conflicts"
			                " with \"%s\" at \"%s\".\n",
			        path, conflict->get_fq_name()->flatten()->c_str(),
			        conflict->get_path()->c_str());

			throw UserError();
		}
	} else {
		ctx.parent = toplevel;
	}
	
	ns->path = new String(path);
	ctx.parent->add_import(ns, ctx.filename);
}

void NameSpace::import_all()
{
	if (!path)
		return;
	
	DIR *dir = opendir(path->c_str());
	if (!dir) {
		fprintf(stderr, "Cannot open directory \"%s\".\n", path->c_str());
		throw UserError();
	}
	
	struct dirent ent, *entp;
	while (true) {
		// readdir_r is buggy on osx 10.2, and will fail if errno is
		// non-zero.
		errno = 0;
		
		if (readdir_r(dir, &ent, &entp)) {
			fprintf(stderr, "1 Cannot readdir on \"%s\": %s.\n",
			        path->c_str(), strerror(errno));

			closedir(dir);
			throw UserError();
		}
		
		if (!entp)
			break;
		
		// Ignore ".", "..", and ".self".
		if (ent.d_name[0] == '.')
			continue;
		
		try {
			char *under = strchr(ent.d_name, '_');
			if (!under) {
				fprintf(stderr, "\"%s\" is not a valid namespace "
				                "(bad member \"%s\").\n",
				        path->c_str(), ent.d_name);

				throw UserError();
			}
			
			String *s = new String(under + 1);
			lookup(s);
		}
		
		catch (SymbolNotFound) {
			fprintf(stderr, "\"%s\" is not a valid namespace "
			                "(member \"%s\" disappeared).\n",
			        path->c_str(), ent.d_name);

			closedir(dir);
			throw UserError();
		}
	}
	
	closedir(dir);
}

void NameSpace::import_all_recursive()
{
	import_all();
	
	for (const_iterator i = begin(); i != end(); ++i) {
		Symbol *sym = (*i).second;
		NameSpace *ns = dynamic_cast<NameSpace *>(sym);
		if (ns)
			ns->import_all_recursive();
	}
}
