// idlcomp/languages/c++/server.cc -- C++ server-side stubs
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <fstream>
#include <errno.h>

#include <idlc.h>
#include <cdl.h>
#include <targets.h>
#include "c++.h"

using std::ofstream;

static void output_ifaceptrs(Class *cla, ostream &f)
{
	f << "// " << *cla->get_fq_name()->flatten("::") << "\n"
	     "// This is a C++ server class mixin generated by idlc.\n"
	     "// Do not modify this file directly.\n"
	     "// This file provides data and methods used by the object system\n"
	     "// to connect this class to its interface(s).  Include it inside\n"
	     "// the class definition as \"public:\", and call the init_iface()\n"
	     "// method from your constructor(s).\n\n";

	
	Interface *conc = cla->concrete_iface;
	assert(conc->get_num_chains() > 0);
	
	f << "struct IDL_ifaceptr {\n\t";
	
	for (int i = 0; i < conc->get_num_chains(); i++) {
		Interface *iface = conc->get_chain_head(i);

		// Skip the "fake" aggregation interface on the first chain.
		if (i == 0)
			iface = *iface->supers_begin();

		cpp_output_name(f, iface, "_i_");
		f << " p" << i << ";\n\t";
	}
		
	f << **cla->name << " *priv;\n"
	     "\tIDL_ifaceptr();\n"
	     "};\n\n";
}

static void output_classptr(Class *cla, Interface *iface, ostream &f)
{
	int chain = cla->concrete_iface->super_to_chain(iface);

	f << "static " << **cla->name << " *classptr(";
	cpp_output_name(f, iface);
	f << " ref)\n"
	     "{\n"
	     "\tif (ref && static_cast<const void *>(ref._ptr->info) ==\n"
	     "\t           static_cast<const void *>(&IDL_info.p" << chain << "))\n"
	     "\t{\n"
	     "\t\tuintptr_t ptr = reinterpret_cast<uintptr_t>(ref._ptr);\n"
	     "\t\tptr -= " << chain * target->pointer_size << ";\n"
	     "\t\treturn reinterpret_cast<IDL_ifaceptr *>(ref._ptr)->priv;\n"
	     "\t}\n\n"
	     "\treturn NULL;\n"
	     "}\n\n";
}

static void output_convfunc(Class *cla, Interface *iface, ostream &f)
{
	f << "operator ";

	cpp_output_name(f, iface);

	f << "()\n"
	     "{\n"
	     "\treturn ";

	cpp_output_name(f, iface);

	int chain = cla->concrete_iface->super_to_chain(iface);

	f << "(reinterpret_cast< ";
	cpp_output_name(f, iface, "_i_");
	f << " *>(&IDL_ref.p" << chain << "));\n"
	     "}\n\n";
}

static void output_info_type(Class *cla, ostream &f)
{
	Interface *conc = cla->concrete_iface;
	f << "struct IDL_info_type {\n";
	
	for (int i = 0; i < conc->get_num_chains(); i++) {
		Interface *iface = conc->get_chain_head(i);
		
		// Skip the "fake" aggregation interface on the first chain.
		if (i == 0)
			iface = *iface->supers_begin();

		f << '\t';
		cpp_output_name(f, iface, "_i_");
		f << "::info_type p" << i << ";\n";
	}

	f << "};\n\n"
	     "static IDL_info_type IDL_info;\n\n";
}

static void output_mixin(Class *cla, string &filename)
{
	ofstream f;
	
	f.open(filename.c_str());
	if (!f.is_open()) {
		fprintf(stderr, "Could not create output file \"%s\": %s.\n",
		        filename.c_str(), strerror(errno));

		throw UserError();
	}
	
	output_ifaceptrs(cla, f);
	output_info_type(cla, f);

	f << "// The classptr method(s) can be used to get a class pointer from\n"
	     "// an interface reference.  If the interface reference does\n"
	     "// not refer to an instance of the class, it returns NULL.\n\n";

	for (Class::ifaces_iterator i = cla->ifaces_begin();
	     i != cla->ifaces_end(); ++i)
	{
		output_classptr(cla, *i, f);
	}
	
	f << "// Used by auto-generated stubs which know that the interface is of\n"
	     "// the proper class.  Do not call this from user code.\n\n"
	     "static " << **cla->name << " *_classptr_internal(void *ptr)\n"
	     "{\n"
	     "\tIDL_ifaceptr *wrap = static_cast<IDL_ifaceptr *>(ptr);\n"
	     "\treturn wrap->priv;\n"
	     "}\n\n";

	f << "IDL_ifaceptr IDL_ref;\n\n";

	f << "// The implicit conversion function(s) below can be used to convert\n"
	     "// a class reference to an interface reference.  Note that if you\n"
	     "// have a pointer, rather than a reference, you will need to dereference\n"
	     "// it prior to casting to the desired interface type.\n\n";
	
	for (Class::ifaces_iterator i = cla->ifaces_begin();
	     i != cla->ifaces_end(); ++i)
	{
		output_convfunc(cla, *i, f);
	}
	
	f << "// This method must be called prior to using any interface references\n"
	  << "// to this object.  If a method is called through an interface pointer\n"
	  << "// before this is done, a memory fault is likely.\n\n"
	  << "void init_iface()\n"
	  << "{\n"
	  << "\tIDL_ref.priv = this;\n"
	  << "}\n";
}

struct callback_data {
	Class *cla;
	ostream &f;
};

static void output_method_wrapper(Class *cla, Interface *iface,
                                  Method *meth, Class::MethodInfo *mi,
                                  bool copy, ostream &f)
{
	const String *implname_idlns, *implname_scope;
	
	if (mi) {
		implname_idlns = mi->implname->flatten("_IDLNS_");
		implname_scope = mi->implname->flatten("::");
	} else {
		implname_idlns = implname_scope = meth->name;
		assert(!copy);
	}
	
	Indent indent = { 1, implname_idlns->length() + 6 };
	
	f << "\tvoid ";
	
	if (copy)
		f << "_copy_";
	
	f << *implname_idlns << "(void *_this";
	
	for (Method::entries_iterator i = meth->entries_begin();
	     i != meth->entries_end(); ++i)
	{
		f << ",\n" << indent;
		cpp_output_one_param(f, *i, true, copy);
	}
	
	const String *classname = cla->get_fq_name()->flatten("::");
	
	f << ")\n"
	     "\t{\n";

	int chain = cla->concrete_iface->super_to_chain(iface);
	if (chain) {
		f << "\t\t_this = reinterpret_cast<void *>"
		     "(reinterpret_cast<uintptr_t>(_this) - " 
		  << chain * target->pointer_size << ");\n";
	}

	f << "\t\t::" << *classname << " *_ptr = ::" << *classname
	  << "::_classptr_internal(_this);\n"
	     "\t\t_ptr->" << *implname_scope << '(';
	
	indent = (Indent){ 2, implname_scope->length() + 7 };

	for (Method::entries_iterator i = meth->entries_begin();
	     i != meth->entries_end(); ++i)
	{
		if (i != meth->entries_begin())
			f << ",\n" << indent;

		Param *p = *i;
		bool copy_this_param = false;
		if (copy) {
			Class::ParamInfo *pi = mi->get_param(p);
			if (pi && pi->copy)
				copy_this_param = true;
		}

		// FIXME: copy non-arrays

		f << **p->name;
		
		if (copy_this_param && p->is_array())
			f << ".copy()";
	}
	
	f << ");\n"
	     "\t}\n\n";
}

// Should be static, but GCC won't accept it as a template parameter
// even though the template is only used from this file (and thus
// won't be instantiated elsewhere (unless some template consolidation
// mechanism puts them all in one file or something, but then an
// exception could be made for templates with static parameters,
// as they can be placed in their original files with no chance
// of a duplicate)).

void cpp_output_iface_method_wrappers(Interface *super, void *arg)
{
	callback_data *data = static_cast<callback_data *>(arg);
	
	for (Interface::methods_iterator i = super->methods_begin();
	     i != super->methods_end(); ++i)
	{
		Method *m = *i;
		Class::MethodInfo *mi = data->cla->get_method(m);

		output_method_wrapper(data->cla, super, m, mi, false, data->f);
		
		if (mi && mi->copy_params)
			output_method_wrapper(data->cla, super, m, mi, true, data->f);
	}
}

struct Context {
	Class *cla;
	ostream &f;
	Indent indent;
	int chain;
	Interface *cur;
};

void cpp_output_iface_table_entry(Interface *super, void *arg)
{
	Context &ctx(*static_cast<Context *>(arg));
	
	ctx.f << ctx.indent << "{\n";
	ctx.indent.indent_level++;

	ctx.f << ctx.indent;
	cpp_output_name(ctx.f, super);
	
	ctx.f << "_ns::_guid.l,\n"
	      << ctx.indent 
	      << (ctx.cla->concrete_iface->super_to_chain(super) - ctx.chain) *
	         target->pointer_size << '\n';

	ctx.indent.indent_level--;
	ctx.f << ctx.indent << "},\n";
}

void cpp_output_method_table_entries(Interface *super, void *arg)
{
	Context &ctx(*static_cast<Context *>(arg));
	
	for (Interface::methods_iterator i = super->methods_begin();
	     i != super->methods_end(); ++i)
	{
		Method *meth = *i;
		Class::MethodInfo *mi = ctx.cla->get_method(meth);
		const String *implname;
	
		if (mi)
			implname = mi->implname->flatten("::");
		else
			implname = meth->name;

		ctx.f << ctx.indent << "IDL_Server_"
		      <<  *ctx.cla->get_fq_name()->flatten("_IDLNS_")
		      << "::" << *implname << ",\n";
	}
}

static void output_info_type_init_rec(Context &ctx)
{
	Interface *iface = ctx.cur;
	
	ctx.f << ctx.indent << "{ // " << *iface->get_fq_name()->flatten() << '\n';
	ctx.indent.indent_level++;
	
	if (!iface->supers_empty()) {
		ctx.cur = *iface->supers_begin();
		output_info_type_init_rec(ctx);
	} else {
		// Concrete pointer adjustment
		ctx.f << ctx.indent << '-' 
		      << ctx.chain * target->pointer_size << ",\n";
		
		// Pointer to concrete downcast table
		ctx.f << ctx.indent << "&IDL_info.p0.iface_table.";
		
		Interface *iface = *ctx.cla->concrete_iface->supers_begin();
		ctx.f << *iface->get_fq_name()->flatten("_IDLNS_") << ",\n";
	}

	// Interface table for downcasts
	ctx.f << ctx.indent << "{\n";
	ctx.indent.indent_level++;
	
	cpp_output_iface_table_entry(iface, &ctx);
	iface->for_each_super<cpp_output_iface_table_entry>(&ctx);
	
	ctx.f << ctx.indent << "NULL\n";
	
	ctx.indent.indent_level--;
	ctx.f << ctx.indent << "},\n";

	// Method table
	
	ctx.f << ctx.indent << "{\n";
	ctx.indent.indent_level++;
	
	cpp_output_method_table_entries(iface, &ctx);
	iface->for_each_super<cpp_output_method_table_entries>(&ctx);
	
	ctx.indent.indent_level--;
	ctx.f << ctx.indent << "},\n";

	ctx.indent.indent_level--;
	ctx.f << ctx.indent << "},\n";
}

static void output_info_type_init(Class *cla, ostream &f)
{
	Indent indent = (Indent){1, 0};
	
	// The struct name is enclosed in parentheses to disambiguate the
	// leading :: from being a continuation of the type name.
	
	cpp_output_name(f, cla);
	f << "::IDL_info_type (";
	cpp_output_name(f, cla);
	f << "::IDL_info) =\n"
	     "{\n";
	
	Context ctx = { cla, f };
	ctx.indent.indent_level = 1;
	
	for (ctx.chain = 0; ctx.chain < cla->concrete_iface->get_num_chains();
	     ctx.chain++)
	{
		ctx.cur = cla->concrete_iface->get_chain_head(ctx.chain);
		
		// Skip the "fake" aggregation interface on the first chain.
		if (ctx.chain == 0)
			ctx.cur = *ctx.cur->supers_begin();
		
		output_info_type_init_rec(ctx);
	}

	assert(ctx.indent.indent_level == 1);
	f << "};\n\n";
}

static void output_footer(Class *cla, ostream &f)
{
	// Create an arbitrary, unique name to stick the method wrappers in.
	f << "\n\nnamespace IDL_Server_" << *cla->get_fq_name()->flatten("_IDLNS_")
	  << " {\n";

	callback_data data = { cla, f };

	cla->concrete_iface->for_each_super
		<cpp_output_iface_method_wrappers>(&data);

	f << "};\n\n";
	
	output_info_type_init(cla, f);

	cpp_output_name(f, cla);
	f << "::IDL_ifaceptr::IDL_ifaceptr() :";
	
	for (int i = 0; i < cla->concrete_iface->get_num_chains(); i++) {
		if (i != 0)
			f << ',';
	
		f << "\np" << i << "(&IDL_info.p" << i << ")";
	}
	
	f << "\n{\n"
	     "}\n";
}

void CPPBinding::output_server(UserNameSpace *ns, const char *dir)
{
	ofstream footer;

	string footer_name(dir);
	footer_name += "/footer.cc";

	if (makedep) {
		printf("%s: %s\n", footer_name.c_str(), dir);
	} else {	
		footer.open(footer_name.c_str());
		if (!footer.is_open()) {
			fprintf(stderr, "Could not create output file \"%s\": %s.\n",
			        footer_name.c_str(), strerror(errno));
	
			throw UserError();
		}
	
		footer << "// This is a C++ server binding generated by idlc.\n"
		          "// Do not modify this file directly.\n"
		          "// Include this file from exactly one file that has access to\n"
		          "// the class and interface declarations (and preferably, with\n"
		          "// access to to inlineable methods as well).";
	}
	
	for (list<ClassRef>::iterator i = classes.begin(); i != classes.end(); ++i) {
		Class *cla = *i;
	
		string filename(dir);
		filename += '/';
		filename += *cla->get_ns()->get_fq_name()->flatten("/");
		
		makepath(filename.c_str(), false);
		
		filename += '/';
		filename += **cla->name;
		filename += ".h";

		if (makedep) {
			printf("%s: %s\n", filename.c_str(), dir);
		} else {
			output_mixin(cla, filename);
			output_footer(cla, footer);
		}
	}
}
