#ifndef _KERNEL_TIME_H
#define _KERNEL_TIME_H

#include <kernel/kernel.h>
#include <System/Time.h>
#include <util/heap.h>
#include <util/spinlock.h>

namespace Time {
	static inline void fix_time_add(System::Time::Time &time)
	{
		if (time.nanos >= 1000000000) {
			time.nanos -= 1000000000;
			time.seconds++;
		}
	}

	static inline void fix_time_sub(System::Time::Time &time)
	{
		if ((int32_t)time.nanos < 0) {
			time.nanos += 1000000000;
			time.seconds--;
		}
	}
}

static inline long operator < (const System::Time::Time &left,
                               const System::Time::Time &right)
{
	return left.seconds - right.seconds < 0 ||
	       (left.seconds == right.seconds &&
	        left.nanos < right.nanos);
}

static inline long operator == (const System::Time::Time &left,
                                const System::Time::Time &right)
{
	return left.seconds == right.seconds &&
	       left.nanos == right.nanos;
}

static inline long operator != (const System::Time::Time &left,
                                const System::Time::Time &right)
{
	return left.seconds != right.seconds ||
	       left.nanos != right.nanos;
}

static inline System::Time::Time operator +(const System::Time::Time &left,
                                            const System::Time::Time &right)
{
	System::Time::Time ret;

	ret.seconds = left.seconds + right.seconds;
	ret.nanos = left.nanos + right.nanos;
	
	Time::fix_time_add(ret);
	return ret;
}

static inline System::Time::Time &operator +=(System::Time::Time &left,
                                              const System::Time::Time &right)
{
	left.seconds += right.seconds;
	left.nanos += right.nanos;
	
	Time::fix_time_add(left);
	return left;
}

static inline System::Time::Time operator -(const System::Time::Time &left,
                                            const System::Time::Time &right)
{
	System::Time::Time ret;

	ret.seconds = left.seconds - right.seconds;
	ret.nanos = left.nanos - right.nanos;
	
	Time::fix_time_sub(ret);
	return ret;
}

static inline System::Time::Time &operator -=(System::Time::Time &left,
                                              const System::Time::Time &right)
{
	left.seconds -= right.seconds;
	left.nanos -= right.nanos;
	
	Time::fix_time_sub(left);
	return left;
}

static inline System::Time::Time operator +(const System::Time::Time &left,
                                            uint32_t right)
{
	System::Time::Time ret;

	ret.seconds = left.seconds;
	ret.nanos = left.nanos + right;
	
	Time::fix_time_add(ret);
	return ret;
}

static inline System::Time::Time &operator +=(System::Time::Time &left,
                                              uint32_t right)
{
	left.nanos += right;
	
	Time::fix_time_add(left);
	return left;
}

static inline System::Time::Time operator -(const System::Time::Time &left,
                                            uint32_t right)
{
	System::Time::Time ret;

	ret.seconds = left.seconds;
	ret.nanos = left.nanos - right;
	
	Time::fix_time_sub(ret);
	return ret;
}

static inline System::Time::Time &operator -=(System::Time::Time &left,
                                              uint32_t right)
{
	left.nanos -= right;
	
	Time::fix_time_sub(left);
	return left;
}

namespace Time {
	typedef System::Time::Clock Clock;
	typedef System::Time::Time Time;
	typedef System::Time::ITime ITime;
	typedef System::Time::Timer Timer;
	typedef System::Time::ITimer ITimer;

	using namespace Lock;
	struct TimerEntry;

	class TimerMux {
		Util::Heap<TimerEntry> heap;
		SpinLock lock;
		Clock clock;
		Timer parent;

	public:
		TimerMux(Clock CLOCK, Timer PARENT) :
		clock(CLOCK), parent(PARENT)
		{
		}
	
		void arm(TimerEntry *entry, Time expiry);
		void disarm(TimerEntry *entry);
		void run();
	};

	struct TimerEntry {
		Util::Heap<TimerEntry>::Node heap_node;
		Time expiry;
		TimerMux *mux;
		
		#include _KERNEL_SERVER(time/timer/Time/TimerEntry.h)

		TimerEntry(TimerMux *MUX = NULL) : mux(MUX)
		{
			init_iface();
			
			expiry.seconds = 0;
			expiry.nanos = 0;
		}

		virtual ~TimerEntry()
		{
		}

		void arm(Time expiry)
		{
			mux->arm(this, expiry);
		}
		
		void disarm()
		{
			mux->disarm(this);
		}
		
		void get_expiry(Time *ret, uint8_t *was_armed)
		{
			if (!ret || !was_armed) {
				// FIXME: throw exception
				return;
			}
			
			if (is_armed()) {
				*ret = expiry;
				*was_armed = 1;
			} else {
				*was_armed = 0;
			}
		}
	
		long operator < (TimerEntry &other)
		{
			return expiry < other.expiry;
		}
		
		bool is_armed()
		{
			return heap_node.is_on_heap();
		}
		
		void set_action(System::Events::Event *event);
		
		// This can be overridden in low-level code (specifically
		// the kernel) to execute the action synchronously,
		// inside the timer critical section.  Most code should
		// (and non-C++ code must) use set_action instead.
		
		virtual void action();
	};
}

#endif
