// Generic radix trees, which supports mapping integer keys to objects.
//
// Written by Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// This software is in the public domain.

#ifndef _UTIL_RADIX_H
#define _UTIL_RADIX_H

#include <assert.h>
#include <stddef.h>
#include <stdint.h>

#include <lowlevel/bitops.h>

namespace Util {
	// Each leaf node in the tree will contain 2**radix_bits nodes.
	// Each directory node will contain 2**dir_bits pointers.
	// Key must be an unsigned integer.

	template <typename T, typename Key, int radix_bits,
	          int dir_bits = radix_bits>
	class RadixTree {
		enum {
			node_size = 1 << radix_bits,
			dir_size = 1 << dir_bits,
			key_bits = sizeof(Key) * 8,
		};
		
		struct Node {
			const int shift;
			
		protected:
			Node(int SHIFT) : shift(SHIFT)
			{
			}
		};
		
		struct LeafNode : public Node {
			T nodes[node_size];
			
			LeafNode() : Node(0)
			{
				memset(nodes, 0, sizeof(nodes));
			}
		};
		
		struct DirNode : public Node {
			Node *ptrs[dir_size];

			DirNode(int shift) : Node(shift)
			{
				memset(ptrs, 0, sizeof(ptrs));
			}
		};
	
		Node *toplevel;
		
		static int key_to_dir_offset(Key key, int shift)
		{
			return (key >> shift) & (dir_size - 1);
		}

		static uint key_to_offset(Key key)
		{
			return key & (node_size - 1);
		}
		
	public:
	 	// The tree depth is contained in each node, and nodes are never
	 	// removed from the tree, so lockless lookups are possible. 
	 	// Synchronization is required when add == true.
		
		T *lookup(Key key, bool add = false)
		{
			Node *node = toplevel;
			int next_shift = node->shift ? node->shift + dir_bits : radix_bits;
			
			while (unlikely(next_shift < key_bits && (key >> next_shift)))
			{
				if (!add)
					return NULL;
				
				DirNode *dn = new DirNode(next_shift);
				dn->ptrs[0] = node;
				toplevel = dn;
				next_shift += dir_bits;
			}
			
			node = toplevel;
			
			while (node->shift > 0) {
				int off = key_to_dir_offset(key, node->shift);
				DirNode *dn = static_cast<DirNode *>(node);
				node = dn->ptrs[off];
				
				if (!node) {
					if (!add)
						return NULL;
				
					if (dn->shift == radix_bits)
						node = new LeafNode;
					else
						node = new DirNode(dn->shift == radix_bits ?
						                   0 : dn->shift - dir_bits);

					dn->ptrs[off] = node;
				}
			}
			
			LeafNode *ln = static_cast<LeafNode *>(node);
			return &ln->nodes[key_to_offset(key)];
		}
		
		RadixTree()
		{
			// OPT: start with a NULL toplevel.  Besides avoiding wasting
			// memory for empty trees, it avoids the need for a stack of
			// trees down the ptr[0] path if there are no small keys.
			
			toplevel = new LeafNode;
		}
	};

}

#endif
