// util/rbtree.h -- A red/black tree implementation
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#ifndef _UTIL_RBTREE_H
#define _UTIL_RBTREE_H

#include <assert.h>
#include <stddef.h>
#include <stdint.h>

namespace Util {
	// T must have an RBTree<T, NodeV, KeyV>::Node member called rbtree_node.
	// NodeV < NodeV, NodeV < KeyV, and NodeV > KeyV must be supported.
	//
	// Using NodeV != KeyV allows things such as having NodeV represent a
	// range of values and KeyV a single value, causing find() to search
	// for the node in whose range KeyV falls.  It is an error to add
	// nodes that are not well ordered.

	template <typename T, typename NodeV, typename KeyV>
	class RBTree {
	public:
		struct Node {
			Node *parent, *left, *right;
			NodeV value;
			bool red;
			
			// A pointer to the parent's left or right pointer
			// (whichever side this node is on).  This is also
			// used for sanity checks (it is NULL for nodes not
			// on a tree).
			
			Node **parentlink;
			
			Node()
			{
				parent = left = right = NULL;
				parentlink = NULL;
			}
			
			bool is_on_rbtree()
			{
				return parentlink != NULL;
			}
		};
	
	private:
		Node *top;
	
		T *node_to_type(Node *n)
		{
			return reinterpret_cast<T *>(reinterpret_cast<uintptr_t>(n) - 
			                             offsetof(T, rbtree_node));
		}
		
		void assert_rb_condition(Node *left, Node *right)
		{
			assert(left->value < right->value);
			assert(!(right->value < left->value));
			assert(!left->red || !right->red);
		}
		
		void assert_well_ordered(Node *left, NodeV right)
		{
			assert(left->value < right);
			assert(!(left->value > right));
			assert(!left->red || !right->red);
		}
		
		Node *find_node(Node *n, NodeV &val, Node ***link);
		Node *find_node_key(KeyV val, bool exact = true);
		
		bool node_is_left(Node *n)
		{
			assert(n->parentlink == &n->parent->left ||
			       n->parentlink == &n->parent->right);
		
			return n->parentlink == &n->parent->left;
		}
		
		Node *sibling(Node *n)
		{
			if (node_is_left(n))
				return n->parent->right;

			return n->parent->left;
		}
		
		Node *uncle(Node *n)
		{
			return sibling(n->parent);
		}
		
		void rotate_left(Node *n);
		void rotate_right(Node *n);
		
		// B must have no right child and have A as an ancestor.
		void swap(Node *a, Node *b);

		bool red(Node *n)
		{
			return n && n->red;
		}
		
public:
		RBTree()
		{
			top = NULL;
		}
		
		bool empty()
		{
			return top == NULL;
		}
		
		T *find(KeyV value)
		{
			Node *n = find_node_key(value);
			
			if (n)
				return node_to_type(n);
				
			return NULL;
		}
		
		T *find_nearest(KeyV value)
		{
			Node *n = find_node_key(value, false);
			
			if (n)
				return node_to_type(n);
			
			// Should only happen on an empty tree
			return NULL;
		}

		void add(T *t);
		void del(T *t);
	};
	
	template <typename T, typename NodeV, typename KeyV>
	typename RBTree<T, NodeV, KeyV>::Node *
	RBTree<T, NodeV, KeyV>::find_node(Node *n, NodeV &val, Node ***link)
	{
		if (link) {
			if (n == top)
				*link = &top;
			else
				*link = n->parentlink;
		}
		
		if (!top)
			return NULL;
		
		while (true) {
			assert(n);
			Node *next;
		
			if (n->value < val) {
				next = n->right;
				
				if (next) {
					assert_rb_condition(n, next);
				} else {
					if (link)
						*link = &n->right;

					break;
				}
			} else {
				// This assert detects duplicate nodes, but not
				// overlapping ranges (or otherwise non-well-ordered
				// values).
				
				assert(val < n->value);
				next = n->left;
				
				if (next) {
					assert_rb_condition(next, n);
				} else {
					if (link)
						*link = &n->left;

					break;
				}
			}
			
			n = next;
		}
		
		return n;
	}

	template <typename T, typename NodeV, typename KeyV>
	typename RBTree<T, NodeV, KeyV>::Node *
	RBTree<T, NodeV, KeyV>::find_node_key(KeyV val, bool exact)
	{
		Node *n = top;
		
		while (n) {
			Node *next;
		
			if (n->value < val) {
				next = n->right;
				
				if (next)
					assert_rb_condition(n, next);
			} else if (n->value > val) {
				next = n->left;

				if (next)
					assert_rb_condition(next, n);
			} else {
				break;
			}
			
			if (!next && !exact)
				return n;
			
			n = next;
		}
		
		return n;
	}
	
	template <typename T, typename NodeV, typename KeyV>
	void RBTree<T, NodeV, KeyV>::rotate_left(Node *n)
	{
		Node *new_top = n->right;
		
		assert(*n->parentlink == n);
		*n->parentlink = new_top;
		
		assert(new_top->parent == n);
		assert(new_top->parentlink == &n->right);

		new_top->parent = n->parent;
		new_top->parentlink = n->parentlink;
		
		n->parent = new_top;
		n->parentlink = &new_top->left;
		
		n->right = new_top->left;
		new_top->left = n;
		
		if (n->right) {
			assert(n->right->parent == new_top);
			assert(n->right->parentlink == &new_top->left);
		
			n->right->parent = n;
			n->right->parentlink = &n->right;
		}
	}

	template <typename T, typename NodeV, typename KeyV>
	void RBTree<T, NodeV, KeyV>::rotate_right(Node *n)
	{
		Node *new_top = n->left;
		
		assert(*n->parentlink == n);
		*n->parentlink = new_top;
		
		assert(new_top->parent == n);
		assert(new_top->parentlink == &n->left);

		new_top->parent = n->parent;
		new_top->parentlink = n->parentlink;
		
		n->parent = new_top;
		n->parentlink = &new_top->right;
		
		n->left = new_top->right;
		new_top->right = n;
		
		if (n->left) {
			assert(n->left->parent == new_top);
			assert(n->left->parentlink == &new_top->right);

			n->left->parent = n;
			n->left->parentlink = &n->left;
		}
	}
	
	// B must have no right child and have A as an ancestor.
	template <typename T, typename NodeV, typename KeyV>
	void RBTree<T, NodeV, KeyV>::swap(Node *a, Node *b)
	{
		Node *bparent = b->parent;
		Node **bparentlink = b->parentlink;
		
		assert(!b->right);
		assert(a->left || a->right);
		
		b->parent = a->parent;
		b->parentlink = a->parentlink;
		
		if (bparent == a) {
			a->parent = b;
			a->parentlink = &b->left;
		} else {
			a->parent = bparent;
			a->parentlink = bparentlink;
		}
		
		assert(a->parent != a);
		assert(b->parent != b);
	
		Node *bleft = b->left;
		b->left = a->left;
		a->left = bleft;

		b->right = a->right;
		a->right = NULL;

		*a->parentlink = a;
		*b->parentlink = b;
		
		assert(a->parent != a);
		assert(b->parent != b);
		
		bool bred = b->red;
		b->red = a->red;
		a->red = bred;
		
		if (a->left) {
			a->left->parent = a;
			a->left->parentlink = &a->left;
		}

		if (b->right) {
			b->right->parent = b;
			b->right->parentlink = &b->right;
		}

		assert(b->left);
		b->left->parent = b;
		b->left->parentlink = &b->left;
	}

	template <typename T, typename NodeV, typename KeyV>
	void RBTree<T, NodeV, KeyV>::add(T *t)
	{
		Node *n = &t->rbtree_node;
		assert(!n->is_on_rbtree());
		
		Node *insert_at = find_node(top, n->value, &n->parentlink);
		
		assert(insert_at || !top);
		*n->parentlink = n;
		n->parent = insert_at;
		n->left = n->right = NULL;
		
	repeat:
		assert(n->parentlink);
		assert(*n->parentlink == n);

		if (!n->parent) {
			n->red = false;
			return;
		}
		
		assert(n->parent->value < n->value != n->value < n->parent->value);
		n->red = true;

		if (!n->parent->red)
			return;

		Node *unc = uncle(n);
		if (red(unc)) {
			assert(!n->parent->parent->red);
			n->parent->red = unc->red = false;
			n = n->parent->parent;
			goto repeat;
		}
		
		if (node_is_left(n)) {
			if (!node_is_left(n->parent)) {
				rotate_right(n->parent);
				n = n->right;
			}
		} else {
			if (node_is_left(n->parent)) {
				rotate_left(n->parent);
				n = n->left;
			}
		}
		
		assert(n->parent->red);
		assert(!red(uncle(n)));
		assert(!n->parent->parent->red);
		
		n->parent->red = false;
		n->parent->parent->red = true;
		
		if (node_is_left(n)) {
			assert(node_is_left(n->parent));
			rotate_right(n->parent->parent);
		} else {
			assert(!node_is_left(n->parent));
			rotate_left(n->parent->parent);
		}
	}

	template <typename T, typename NodeV, typename KeyV>
	void RBTree<T, NodeV, KeyV>::del(T *t)
	{
		Node *n = &t->rbtree_node;
		assert(*n->parentlink == n);

		if (n->left && n->right) {
			Node *highest_on_left = find_node(n->left, n->value, NULL);
			assert(!highest_on_left->right);
			swap(n, highest_on_left);
			assert(!n->right);
		}

		Node *parent = n->parent;
		Node *child = n->left ? n->left : n->right;
		
		if (child) {
			assert(child->parent == n);

			child->parent = n->parent;
			child->parentlink = n->parentlink;
			*child->parentlink = child;
			assert(child != parent);
		} else {
			*n->parentlink = NULL;
		}
		
		n->parentlink = NULL;
		
		if (n->red)
			return;
		
		n = child;

		if (red(n)) {
			n->red = false;
			return;
		}

	repeat:
		if (n == top) {
			assert(!red(n));
			return;
		}
		
		Node *sib;
		
		if (n)
			sib = sibling(n);
		else
			sib = parent->left ? parent->left : parent->right;
		
		if (sib->red) {
			assert(!parent->red);
			assert(!red(sib->left));
			assert(!red(sib->right));
			
			parent->red = true;
			sib->red = false;

			if (node_is_left(sib)) {
				rotate_right(parent);
				sib = parent->left;
			} else {
				rotate_left(parent);
				sib = parent->right;
			}
			
			if (n)
				assert(sib == sibling(n));
		} else if (!parent->red && !red(sib->left) && !red(sib->right)) {
			sib->red = true;
			assert(n != parent);
			n = parent;
			parent = parent->parent;
			goto repeat;
		}

		assert(!sib->red);
		
		if (!parent->red && !red(sib->left) && !red(sib->right)) {
			sib->red = true;
			parent->red = false;
			return;
		}
		
		if (!red(sib->left) && !red(sib->right)) {
			assert(parent->red);
			sib->red = true;
			parent->red = false;
			return;
		}
		
		if (node_is_left(sib) && !red(sib->left) && red(sib->right)) {
			sib->red = true;
			sib->right->red = false;
			rotate_left(sib);
			sib = sib->parent;
		} else if (!node_is_left(sib) && !red(sib->right) && red(sib->left)) {
			sib->red = true;
			sib->left->red = false;
			rotate_right(sib);
			sib = sib->parent;
		}
		
		assert(parent == sib->parent);
		assert(!sib->red);
		
		sib->red = parent->red;
		parent->red = false;
		
		if (node_is_left(sib)) {
			assert(sib->left->red);
			sib->left->red = false;
			rotate_right(parent);
		} else {
			assert(sib->right->red);
			sib->right->red = false;
			rotate_left(parent);
		}
	}
}

#endif
