// FIXME: PAE support

#ifndef _ARCH_PAGETABLE_H
#define _ARCH_PAGETABLE_H

#include <lowlevel/misc.h>
#include <lowlevel/atomic.h>

namespace Arch {
	union PTE {
		typedef ulong PhysAddr, VirtAddr;
		typedef PTE DirPTE;
	
		struct {
			PhysAddr Valid:1;
			PhysAddr Writeable:1;  // Low-level write access:
			                       // equivalent to PermWrite && !FaultOnWrite
			PhysAddr User:1;
			PhysAddr WriteThrough:1;
			PhysAddr CacheDisable:1;
			PhysAddr Accessed:1;
			PhysAddr Dirty:1;
			PhysAddr PageAttrTable:1;
			PhysAddr Global:1;
			PhysAddr FaultOnWrite:1;
			PhysAddr PermWrite:1;  // High-level permission-based write access
			PhysAddr Avail:1;
			PhysAddr Addr:20;
		};
		
		PhysAddr raw;

		PTE(PhysAddr init) : raw(init)
		{
		}
		
		PTE() : raw(0)
		{
		}

		operator PhysAddr()
		{
			return raw;
		}

		static uint addr_to_offset(VirtAddr addr, int shift)
		{
			int pages_per_table = page_size / sizeof(PhysAddr);
			return (addr >> shift) & (pages_per_table - 1);
		}
	
		PhysAddr pte_to_addr()
		{
			return raw & 0xfffff000;
		}
	
		static PTE addr_to_pte(PhysAddr phys)
		{
			return phys & 0xfffff000;
		}
		
		// DirPTE only
		void *get_table()
		{
			return static_cast<PTE *>(phys_to_kvirt(pte_to_addr()));
		}
		
		// DirPTE only
		static DirPTE set_table(void *addr)
		{
			return addr_to_pte(kvirt_to_phys(addr)) | 7;
		}
		
		static void flags_to_pte(Mem::PTEFlags flagsin,
		                         Mem::PTEFlags maskin,
		                         PTE &flagsout,
		                         PTE &maskout)
		{
			maskout = 0;
			flagsout = 0;
			
			if (maskin.Valid) {
				maskout.Valid = 1;
				flagsout.Valid = flagsin.Valid;
			}
			
			if (maskin.FaultOnWrite) {
				maskout.FaultOnWrite = 1;
				flagsout.FaultOnWrite = flagsin.FaultOnWrite;
			}
			
			if (maskin.Writeable) {
				maskout.Writeable = 1;
				maskout.PermWrite = 1;
				flagsout.Writeable = flagsin.Writeable;
				flagsout.PermWrite = flagsin.Writeable;
				
				// This must be done here if both mask bits are set,
				// as we may not be going through set_flags in that case.

				if (maskin.FaultOnWrite)
					flagsout.Writeable &= !flagsout.FaultOnWrite;
			}
			
			if (maskin.User) {
				maskout.User = 1;
				flagsout.User = flagsin.User;
			}
		}
		
		PTE set_flags(PTE mask, PTE flags)
		{
			PTE ret, new_pte;
		
			do {
				ret = raw;
				new_pte = (raw & ~mask) | flags;
				new_pte.Writeable &= !new_pte.FaultOnWrite;
			} while (!ll_cmpxchg_long(reinterpret_cast<PhysAddr *>
			                          (this), ret, new_pte));

			return ret;
		}
	
		Mem::PTEFlags pte_to_flags()
		{
			Mem::PTEFlags ret = 0;
			
			ret.Valid = Valid;
			ret.User = User;
			
			if (Valid) {
				ret.Readable = 1;
				ret.Writeable = PermWrite;
				ret.Executable = 1;
				ret.FaultOnWrite = FaultOnWrite;
			}
			
			return ret;
		}
	
		void set_pte(PTE *table, uint offset)
		{
			table[offset] = raw;
		}
		
		PTE xchg_pte(PTE *table, uint offset)
		{
			return ll_xchg_long(reinterpret_cast<PhysAddr *>(&table[offset]), raw);
		}
		
		bool valid_pte()
		{
			return Valid;
		}
		
		bool dirty_pte()
		{
			return Dirty;
		}
		
		enum {
			page_size = Arch::page_size,
			page_shift = Arch::page_shift,
	
			pages_per_table = page_size / sizeof(PhysAddr),
			num_levels = 2,
			shift_per_level = 10,
	
			// All "upper" addresses belong to the kernel.
			
			kmap_start = 512,
			kmap_end = 1023
		};
	};

	static inline void invalidate_tlb_entry(ulong addr)
	{
		ll_invalidate_tlb_entry(addr);
	}
}

#endif
